#include <nds.h>
#include <fat.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>

#include "gui.h"
#include "filehandling.h"
#include "unzip/unzipnds.h"

#define SETTINGNAME "settings.cfg"
#define STATESIZE (0x2000+0x40+8+8+96)

//asm variables
extern u8 g_dipswitch0;			// from cart.s
extern u8 g_dipswitch1;			// from cart.s
extern u8 g_dipswitch2;			// from cart.s
extern int goemon;				// from cart.s
extern u8 g_scaling;			// from gfx.s
extern u8 g_flicker;			// from gfx.s
extern u8 g_gammavalue;			// from gfx.s
extern u32 joycfg;				// io.s

//asm functions
extern void loadcart(int,int);		// cart.s
extern int packState(u32* statePtr);	// cart.s
extern void unpackState(u32* statePtr);	// cart.s
extern void CPU_reset(void);		// cpu.s

extern u8 ROM_Space[];
extern int sleepTime;				// main.c
extern int selectedGame;			// main.c
extern int emuSettings;				// gui.c

extern void loadcart(int,int);		// cart.s
extern void waitVBlank(void);		// main.c
extern void pausVBlank(int count);	// main.c
extern void setEmuSpeed(int speed);	// main.c


int fatAvailable = 0;
configdata cfg;

char currentdir[256];

char *const romFilenames[4][11]={
	{"577l03.10c","577l02.8c","577l01.7c","577l07.3f","577l06.5e","577l05.4e","577l08.4f","577l04.3e","577h09.2f","577h10.5f","577h11.6f"},
	{"577h03.10c","577h02.8c","577h01.7c","577h07.3f","577l06.5e","577h05.4e","577l08.4f","577l04.3e","577h09.2f","577h10.5f","577h11.6f"},
	{"2-ic82.10g","3-ic81.10f","1-ic92.12c","7-1c8.2b","6-ic9.2c","5-ic10.2d","4-ic11.2e","577h09","577h10","577h11"},
	{"621d01.10c","621d02.12c","621a05.6d","621d03.4d","621d04.5d","621a06.5f","621a07.6f","621a08.7f"}
	};
char *const gameNames[]={"gberet","rushatck","gberetb","mrgoemon"};
int gberetFilesizes[]={0x4000,0x4000,0x4000,0x4000,0x4000,0x4000,0x4000,0x4000,0x20,0x100,0x100};
int rushatckFilesizes[]={0x4000,0x4000,0x4000,0x4000,0x4000,0x4000,0x4000,0x4000,0x20,0x100,0x100};
int gberetbFilesizes[]={0x8000,0x4000,0x4000,0x4000,0x4000,0x4000,0x4000,0x20,0x100,0x100};
int mrgoemonFilesizes[]={0x8000,0x8000,0x4000,0x8000,0x8000,0x20,0x100,0x100};
int fileCount[]={11,11,10,8};

//---------------------------------------------------------------------------------
void initFileHandling() {
	strcpy(currentdir,"/");
}


int loadSettings() {
	FILE* myfile;

	if(findFolder("gberet")) {
		pausVBlank(60*3);
		return 1;
	}
	if( (myfile = fopen(SETTINGNAME, "r")) ) {
		fread(&cfg, 1, sizeof(configdata), myfile);
		fclose(myfile);
		if(	!strstr(cfg.magic,"cfg")) {
			drawtext("Error in settings file.", 18, 0);
			pausVBlank(60*3);
			return 1;
		}
	} else {
		drawtext("Couldn't open file:", 18, 0);
		drawtext(SETTINGNAME, 19, 0);
		pausVBlank(60*3);
		return 1;
	}

	g_dipswitch0 = cfg.dipswitch0;
	g_dipswitch1 = cfg.dipswitch1;
	g_dipswitch2 = cfg.dipswitch2;
	g_scaling    = cfg.scaling&1;
	g_flicker    = cfg.flicker&1;
	g_gammavalue = cfg.gammavalue;
	emuSettings  = cfg.emuSettings &~ 0xC0;			// Clear speed setting.
	sleepTime    = cfg.sleepTime;
	joycfg       = (joycfg&~0x400)|((cfg.controller&1)<<10);

	drawtext("Settings loaded.", 18, 0);
	pausVBlank(60*1);
	return 0;
}
void saveSettings() {
	FILE* myfile;

	strcpy(cfg.magic,"cfg");
	cfg.dipswitch0  = g_dipswitch0;
	cfg.dipswitch1  = g_dipswitch1;
	cfg.dipswitch2  = g_dipswitch2;
	cfg.scaling     = g_scaling&1;
	cfg.flicker     = g_flicker&1;
	cfg.gammavalue  = g_gammavalue;
	cfg.emuSettings = emuSettings &~ 0xC0;			// Clear speed setting.
	cfg.sleepTime   = sleepTime;
	cfg.controller  = (joycfg>>10)&1;

	if(findFolder("gberet")) {
		pausVBlank(60*3);
		return;
	}
	if( (myfile = fopen(SETTINGNAME, "w")) ) {
		fwrite(&cfg, 1, sizeof(configdata), myfile);
		fclose(myfile);
	} else {
		drawtext("Couldn't open file:", 18, 0);
		drawtext(SETTINGNAME, 19, 0);
		pausVBlank(60*3);
		return;
	}
	drawtext("Settings saved.", 18, 0);
	pausVBlank(60*1);
	return;
}

//---------------------------------------------------------------------------------
void testFAT() {
	int i, pos=0, oldpos=-1, items=0;
	char * dtable;
	char * strp;
	int pressed = 0;

	if(fatAvailable) {
		dtable = malloc(0x8000);
		if( (items=getDirectory(dtable, currentdir, ".pce")) ) {

			while(1) {
				waitVBlank();
				pressed = getinput();
				pos = getmenupos(pressed, pos, items);
				if(pressed & (KEY_A)) {
					strp = dtable + pos*64;
					if( !(strstr(strp,"~")) ) {
						if(strcmp(strp,"..") == 0) {
							i = strlen(currentdir)-strlen(strrchr(currentdir, '/'));
							currentdir[i] = 0;
							currentdir[i+1] = 0;
							currentdir[0] = '/';
						} else {
							if( (strcmp(currentdir,"/")) )
								strcat(currentdir, "/");
							strncat(currentdir, strp, strlen(strp));
						}
						cls(0);
						items=getDirectory(dtable, currentdir, ".pce");
						pos = 0;
						oldpos = -1;
					}
				}
				if(oldpos != pos) {
					oldpos = pos;
					drawfilelist(dtable, pos, items);
				}
			}
		}
		free(dtable);
	}
}

void loadGame() {
	if(loadRoms(selected,0)) {
		pausVBlank(60*3);
		return;
	}
	selectedGame = selected;
	loadRoms(selectedGame,1);
	if(selectedGame == 3) goemon = 1;
	else goemon = 0;
	setEmuSpeed(0);
	loadcart(0,0);
	if(emuSettings&4)
		loadState();
}

int loadRoms(int game, int doLoad) {
	int i, count;
	u8* romarea = ROM_Space;
	FILE* myfile;
	int* romFilesizes;

	game &= game;
	count = fileCount[game];
	romFilesizes = gberetFilesizes;
	if(game == 1) {
		romFilesizes = rushatckFilesizes;
	} else 	if(game == 2) {
		romFilesizes = gberetbFilesizes;
	} else 	if(game == 3) {
		romFilesizes = mrgoemonFilesizes;
	}

	if(findFolder("gberet"))
		return 1;
	for(i=0;i<count;i++) {
		if( (myfile = fopen(romFilenames[game][i], "r")) ) {
			if(doLoad) fread(romarea, 1, romFilesizes[i], myfile);
			fclose(myfile);
			romarea += romFilesizes[i];
		} else if( !(findFileInZip("gberet.zip", romFilenames[game][i])) ) {
			if(doLoad) loadFileInZip(romarea, "gberet.zip", romFilenames[game][i]);
			romarea += romFilesizes[i];
		} else if( !(findFileInZip("rushatck.zip", romFilenames[game][i])) ) {
			if(doLoad) loadFileInZip(romarea, "rushatck.zip", romFilenames[game][i]);
			romarea += romFilesizes[i];
		} else if( !(findFileInZip("gberetb.zip", romFilenames[game][i])) ) {
			if(doLoad) loadFileInZip(romarea, "gberetb.zip", romFilenames[game][i]);
			romarea += romFilesizes[i];
		} else if( !(findFileInZip("mrgoemon.zip", romFilenames[game][i])) ) {
			if(doLoad) loadFileInZip(romarea, "mrgoemon.zip", romFilenames[game][i]);
			romarea += romFilesizes[i];
		} else {
			drawtext("Couldn't open file:", 18, 0);
			drawtext(romFilenames[game][i], 19, 0);
			return 1;
		}
	}

	return 0;
}

void loadState(void) {
	u32* statePtr;
	FILE* myfile;
	char stateName[32];

	if(findFolder("gberet")) {
		pausVBlank(60*3);
		return;
	}
	strcpy(stateName,gameNames[selectedGame]);
	strcat(stateName,".sta");
	if( (statePtr = malloc(STATESIZE)) ) {
		if( (myfile = fopen(stateName, "r")) ) {
			fread(statePtr, 1, STATESIZE, myfile);
			fclose(myfile);
			unpackState(statePtr);
		}
		free(statePtr);
	} else {
		drawtext("Couldn't alloc mem for state.", 18, 0);
		pausVBlank(60*3);
		return;
	}
	drawtext("Loaded state.", 18, 0);
	pausVBlank(60*1);
}
void saveState(void) {
	u32* statePtr;
	FILE* myfile;
	char stateName[32];

	if(findFolder("gberet")) {
		pausVBlank(60*3);
		return;
	}
	strcpy(stateName,gameNames[selectedGame]);
	strcat(stateName,".sta");
	if( (statePtr = malloc(STATESIZE)) ) {
		packState(statePtr);
		if( (myfile = fopen(stateName, "w")) ) {
			fwrite(statePtr, 1, STATESIZE, myfile);
			fclose(myfile);
		}
		free(statePtr);
	} else {
		drawtext("Couldn't alloc mem for state.", 18, 0);
		pausVBlank(60*3);
		return;
	}
	drawtext("Saved state.", 18, 0);
	pausVBlank(60*1);
}

int findFolder(const char * folderName) {
	int retValue=0;
	if(!chdir(folderName))
		return retValue;
	chdir("/");
	strcpy(currentdir, "data/");
	strncat(currentdir, folderName, 255);
	if(chdir(currentdir)) {
		if(chdir(folderName)) {
			drawtext("Couldn't find folder:", 18, 0);
			drawtext(folderName, 19, 0);
			retValue = 1;
		}
	}
	return retValue;
}

int getDirectory(char * dirtable, const char * dirname, const char * filetype) {
	int i=0;

	char * strp;
	DIR * pdir;
	struct dirent * pent;
	struct stat statbuf;

	memset(dirtable,0,0x8000);
	chdir("/");			// stupid workaround.
	chdir(dirname);

	pdir=opendir(".");

	drawtext(dirname,0,0);
	if (pdir){
		while ((pent=readdir(pdir))!=NULL) {
			strp = dirtable + i*64;
			stat(pent->d_name,&statbuf);
			if(strcmp(".", pent->d_name) == 0)
				continue;
			if(S_ISDIR(statbuf.st_mode)) {
				strncpy(strp, pent->d_name,63);
			}
			else {
				if(strstr(pent->d_name, filetype) == 0)
					continue;
				strcpy(strp, "~");
				strncat(strp, pent->d_name, 63-1);
			}
			i++;
			if(i>511) break;
		}
		closedir(pdir);
		sortDirectory(dirtable, i-1);
	}
	return i;
}

void sortDirectory(char * dirtable, int count) {
	int i, j=0, dirt;
	char temp[64];

	for(j=0;j<count;j++) {
		dirt = 0;
		for(i=0;i<count;i++) {
			if( strcasecmp(dirtable + i*64, dirtable + (i+1)*64) > 0) {
				strcpy(temp, dirtable + i*64);
				strcpy(dirtable + i*64, dirtable + (i+1)*64);
				strcpy(dirtable + (i+1)*64, temp);
				dirt = 1;
			}
		}
		if(!dirt) break;
	}
}

