#include "nds_asm.h"
#include "equates.h"
#include "ARMZ80/z80mac.h"
#include "armsdtmac.h"

	.global GFX_init
	.global GFX_reset
	.global debug_
	.global AGBinput
	.global EMUinput
	.global paletteInit
	.global paletteTxAll
	.global convertPalette
	.global newframe
	.global endframe
	.global gfxstate
	.global g_gammavalue
//	.global oambufferready
	.global g_scaling
	.global g_twitch
	.global g_flicker
	.global g_gfxmask
	.global fpsenabled
	.global FPSValue
	.global vblirqhandler
	.global irqControl
	.global irqPending
	.global ystart
	.global frametotal
	.global scanline
	.global sprBank
	.global sprmemreload

	.global OAM_BUFFER1			;@ sprite scale setup should be moved here!!!

	.arm
	.align 4
	.section .text

	.global rawfont
rawfont:
	.incbin "font.bin"

//	.global rawfontpal
//rawfontpal:
//	.incbin "fontpal.bin"

;@----------------------------------------------------------------------------
//GFX_init:					;@ (called from main.c) only need to call once
;@----------------------------------------------------------------------------
	mov addy,lr

	mov r1,#REG_BASE
	mov r0,#0x0008
	strh r0,[r1,#REG_DISPSTAT]	;@ vblank en


	add r0,r1,#REG_BG0HOFS		;@ DMA0 always goes here
	str r0,[r1,#REG_DMA0DAD]
	ldr r0,=DMA0BUFF+8			;@ DMA0 src=
	str r0,[r1,#REG_DMA0SAD]
	mov r0,#2					;@ 1 word transfer
	strh r0,[r1,#REG_DMA0CNT_L]

	add r2,r1,#REG_IE
	mov r0,#-1
	strh r0,[r2,#2]				;@ stop pending interrupts
//	ldr r0,=irqhandler
	str r0,[r1,#-4]				;@ =AGB_IRQVECT
	ldr r0,=0x1081
	strh r0,[r2]				;@ key(0x1000),vblank(0x0001),timer1(0x0010) enable. (serial interrupt=0x0080)
	mov r0,#1
	strh r0,[r2,#8]				;@ master irq enable

	bx addy
;@----------------------------------------------------------------------------
GFX_reset:					;@ called with CPU reset
;@----------------------------------------------------------------------------
	str lr,[sp,#-4]!

	mov r1,#0
	str r1,wtop

	ldr r0,=gfxstate
	mov r2,#2					;@ 2*4
	bl memset_					;@ clear GFX regs

	mov r0,#1
	strb r0,sprmemreload

	mov r0,#0
	strb r0,ystart

	mov r1,#REG_BASE
	ldr r0,=0x4006
	strh r0,[r1,#REG_BG0CNT]
	ldr r0,=0x428A
	strh r0,[r1,#REG_BG1CNT]

	ldr r0,=0x08F8				;@ start-end
	strh r0,[r1,#REG_WIN0H]
	mov r0,#0x00C0				;@ start-end
	strh r0,[r1,#REG_WIN0V]
	mov r0,#0x0000
	strh r0,[r1,#REG_WINOUT]

	mov r0,#OAM
	mov r1,#0x2a0
	mov r2,#0x100
	bl memset_					;@ no stray sprites please
	ldr r0,=OAM_BUFFER1
	mov r2,#0x200
	bl memset_
	bl BG_init

	bl paletteInit				;@ do palette mapping
	bl paletteTxAll				;@ transfer it
	ldr pc,[sp],#4

;@----------------------------------------------------------------------------
BG_init:					;@ BG tiles
;@----------------------------------------------------------------------------
	ldr r3,=vrombase0			;@ r3 = bitplane 0
	ldr r3,[r3]
	mov r12,#BG_GFX				;@ r12=NDS BG tileset
	add r12,r12,#0x4000			;@ tile ram 1.5
	ldr r1,=0x0F0F0F0F
bgchr1:
	ldr r0,[r3],#4				;@ read 3rd plane
	and r2,r1,r0,lsr#4
	and r0,r1,r0
	orr r0,r2,r0,lsl#4
	str r0,[r12],#4

	tst r12,#0x8000
	beq bgchr1

	ldr r1,=0x10101010
	mov r0,#0
bgchr3:
	mov r2,#16
bgchr2:
	str r0,[r12],#4
	subs r2,r2,#1
	bne bgchr2
	adds r0,r0,r1
	bcc bgchr3

	bx lr

;@----------------------------------------------------------------------------
paletteInit:		;@ r0-r3 modified.
;@ called by ui.c:  void map_palette();
;@----------------------------------------------------------------------------
	stmfd sp!,{r4-r9,lr}
	ldr r8,=prombase			;@ proms
	ldr r8,[r8]
	mov r7,#0xE0
	ldr r6,=MAPPED_RGB
	ldrb r1,g_gammavalue		;@ gamma value = 0 -> 4
	mov r4,#32					;@ Green Beret bgr, r1=R, r2=G, r3=B
nomap:							;@ map 00000000bbgggrrr  ->  0bbbbbgggggrrrrr
	ldrb r9,[r8],#1
	and r0,r9,#0xC0				;@ Blue ready
	bl gprefix
	mov r5,r0

	and r0,r7,r9,lsl#2			;@ Green ready
	bl gprefix
	orr r5,r0,r5,lsl#5

	and r0,r7,r9,lsl#5			;@ Red ready
	bl gprefix
	orr r5,r0,r5,lsl#5

	strh r5,[r6],#2
	subs r4,r4,#1
	bne nomap

	ldmfd sp!,{r4-r9,lr}
	bx lr

;@----------------------------------------------------------------------------
convertPalette:				;@ r0 = destination, r1 = source, r2 = length. r3, r12 modified.
;@ called by main.c:  void convertPalette(void* destination,void* source, int length);
;@----------------------------------------------------------------------------
	stmfd sp!,{r4-r7,lr}
	mov r4,r0					;@ destination
	mov r5,r1					;@ source
	mov r6,r2					;@ length
	ldrb r1,g_gammavalue		;@ gamma value = 0 -> 4
convloop:						;@ map rrrrrrrrbbbbbbbbgggggggg  ->  0bbbbbgggggrrrrr
	ldrb r0,[r5],#1				;@ Red
	bl gammaconvert
	mov r7,r0

	ldrb r0,[r5],#1				;@ Green
	bl gammaconvert
	orr r7,r7,r0,lsl#5

	ldrb r0,[r5],#1				;@ Blue
	bl gammaconvert
	orr r7,r7,r0,lsl#10

	strh r7,[r4],#2
	subs r6,r6,#1
	bhi convloop

	ldmfd sp!,{r4-r7,lr}
	bx lr

;@----------------------------------------------------------------------------
gprefix:
	orr r0,r0,r0,lsr#3
	orr r0,r0,r0,lsr#6
;@----------------------------------------------------------------------------
gammaconvert:	;@ takes value in r0(0-0xFF), gamma in r1(0-4),returns new value in r0=0x1F
;@----------------------------------------------------------------------------
	rsb r2,r0,#0x100
	mul r3,r2,r2
	rsbs r2,r3,#0x10000
	rsb r3,r1,#4
	orr r0,r0,r0,lsl#8
	mul r2,r1,r2
	mla r0,r3,r0,r2
	mov r0,r0,lsr#13

	bx lr
;@----------------------------------------------------------------------------
paletteTxAll:				;@ Called from ui.c
;@----------------------------------------------------------------------------
	stmfd sp!,{r4-r5}

	ldr r2,=prombase			;@ proms
	ldr r2,[r2]
	add r2,r2,#32
	ldr r3,=MAPPED_RGB
	ldr r4,=EMUPALBUFF
	add r5,r4,#512
	mov r1,#256
nomap2:
	ldrb r0,[r2],#1
	mov r0,r0,lsl#1
	ldrh r0,[r3,r0]
	strh r0,[r5],#2
	subs r1,r1,#1
	bne nomap2

	add r3,r3,#32
	mov r1,#256
nomap3:
	ldrb r0,[r2],#1
	mov r0,r0,lsl#1
	ldrh r0,[r3,r0]
	strh r0,[r4],#2
	subs r1,r1,#1
	bne nomap3

	ldmfd sp!,{r4-r5}
	bx lr
;@----------------------------------------------------------------------------
showfps_:		;@ fps output, r0-r3=used.
;@----------------------------------------------------------------------------
	ldrb r0,fpschk
	subs r0,r0,#1
	movmi r0,#59
	strb r0,fpschk
	bxpl lr						;@ End if not 60 frames has passed

	ldrb r0,fpsenabled
	tst r0,#1
	bxeq lr						;@ End if not enabled

	ldr r2,=FPSValue
	ldr r0,[r2]
	cmp r0,#0
	bxeq lr						;@ End if fps==0, to keep it from appearing in the menu
	mov r1,#0
	str r1,[r2]

	mov r1,#100
	swi 0x090000				;@ Division r0/r1, r0=result, r1=remainder.
	add r0,r0,#0x30
	strb r0,fpstext+5
	mov r0,r1
	mov r1,#10
	swi 0x090000				;@ Division r0/r1, r0=result, r1=remainder.
	add r0,r0,#0x30
	strb r0,fpstext+6
	add r1,r1,#0x30
	strb r1,fpstext+7
	

	adr r0,fpstext
	ldr r2,=DEBUGSCREEN
db1:
	ldrb r1,[r0],#1
	orr r1,r1,#0x3100
	strh r1,[r2],#2
	tst r2,#0xE
	bne db1

	bx lr
;@----------------------------------------------------------------------------
fpstext:	.string "FPS:   "
fpsenabled:	.byte 0
fpschk:		.byte 0
			.byte 0
			.byte 0
;@----------------------------------------------------------------------------
debug_:						;@ debug output, r0=val, r1=line, r2=used.
;@----------------------------------------------------------------------------
.if 0
 [ DEBUG
	ldr r2,=DEBUGSCREEN
	add r2,r2,r1,lsl#6
db0
	mov r0,r0,ror#28
	and r1,r0,#0x0f
	cmp r1,#9
	addhi r1,r1,#7
	add r1,r1,#0x30
	orr r1,r1,#0x4100
	strh r1,[r2],#2
	tst r2,#15
	bne db0
 ]
.endif
	bx lr

;@----------------------------------------------------------------------------
vblirqhandler:
;@----------------------------------------------------------------------------
	stmfd sp!,{r4-r8,lr}
	bl showfps_

	ldr r5,=SCROLLBUFF
	mov r4,r5
	ldrb r8,ystart
	add r8,r8,#0x10
	ldr r6,=0x00002B10			;@ was 0x2AAB

	ldrb r0,g_scaling
	cmp r0,#0
	moveq r6,#0
	movne r8,#0x10
	mov r7,r8,lsl#16

	ldrb r0,g_flicker
	ldrb r1,g_twitch
	eors r1,r1,r0
	strb r1,g_twitch
	addne r6,r6,r6,lsl#16

	ldr r2,=SCRL_TMP
	add r3,r2,#0x20
	mov r12,#192
ScrolLoop2:
	ldrb r0,[r2,r8,lsr#3]
	ldrb r1,[r3,r8,lsr#3]
	orr r0,r0,r1,lsl#8
	add r0,r0,r7
	mov r1,r0
	stmia r4!,{r0-r1}
	adds r6,r6,r6,lsl#16
	addcs r7,r7,#0x10000
	adc r8,r8,#1
	subs r12,r12,#1
	bne ScrolLoop2



	mov r6,#REG_BASE
	strh r6,[r6,#REG_DMA0CNT_H]	;@ DMA0 stop

	add r1,r6,#REG_DMA0SAD
	mov r2,r5					;@ setup DMA buffer for scrolling:
	ldmia r2!,{r4-r5}			;@ Read
	add r3,r6,#REG_BG0HOFS		;@ DMA0 always goes here
	stmia r3,{r4-r5}			;@ set 1st value manually, HBL is AFTER 1st line
	ldr r4,=0x96600002			;@ noIRQ hblank 32bit repeat incsrc inc_reloaddst, 2 word
	stmia r1,{r2-r4}			;@ DMA0 go

	add r1,r6,#REG_DMA3SAD

	ldr r2,dmaoambuffer			;@ DMA3 src, OAM transfer:
	mov r3,#OAM					;@ DMA3 dst
	mov r4,#0x84000000			;@ noIRQ 32bit incsrc incdst
	orr r4,r4,#0xc0				;@ 196 words (768 bytes)
	stmia r1,{r2-r4}			;@ DMA3 go

	ldr r2,=EMUPALBUFF			;@ DMA3 src, Palette transfer:
	mov r3,#BG_PALETTE			;@ DMA3 dst
	mov r4,#0x84000000			;@ noIRQ 32bit incsrc incdst
	orr r4,r4,#0x100			;@ 256 words (1024 bytes)
	stmia r1,{r2-r4}			;@ DMA3 go

	mov r0,#0x0013
	ldrb r1,g_gfxmask
	bic r0,r0,r1
	strh r0,[r6,#REG_WININ]

	ldr r0,=scanKeys
	adr lr,exit_vbl
	bx r0
exit_vbl:
	ldmfd sp!,{r4-r8,pc}


;@----------------------------------------------------------------------------
g_scaling:		.byte 1
g_twitch:		.byte 0
g_flicker:		.byte 1
g_gammavalue:	.byte 0
g_gfxmask:		.byte 0
ystart:			.byte 0
				.byte 0,0
;@----------------------------------------------------------------------------
newframe:					;@ called before line 0	(r0-r9 safe to use)
;@----------------------------------------------------------------------------
	mov r0,#0
	str r0,scanline				;@ reset scanline count
	bx lr

;@----------------------------------------------------------------------------
endframe:					;@ called just before screen end (~line 240)	(r0-r2 safe to use)
;@----------------------------------------------------------------------------
	stmfd sp!,{r3-r9,lr}

	bl scrollcopy
	bl bg_finish
;@--------------------------
	bl sprDMA_do
;@--------------------------

//	mrs r5,cpsr
//	orr r1,r5,#0x80				;@--> Disable IRQ.
//	msr cpsr_cf,r1

	ldr r0,dmaoambuffer
	ldr r1,tmpoambuffer
	str r0,tmpoambuffer
	str r1,dmaoambuffer

	mov r0,#1
	str r0,oambufferready

	ldr r0,=windowtop			;@ load wtop, store in wtop+4.......load wtop+8, store in wtop+12
	ldmia r0,{r1-r3}			;@ load with post increment
	stmib r0,{r1-r3}			;@ store with pre increment

//	msr cpsr_cf,r5				;@--> restore mode,Enable IRQ.


	ldmfd sp!,{r3-r9,lr}
	bx lr
	
;@----------------------------------------------------------------------------
scrollcopy:					;@ end of frame...
;@----------------------------------------------------------------------------
	ldr r1,=SCRL_RAM
	ldr r2,=SCRL_TMP
	mov r3,#0x10
scrl_cpy_loop:
	ldr r0,[r1],#4
	str r0,[r2],#4
	subs r3,r3,#1
	bne scrl_cpy_loop
	bx lr

;@----------------------------------------------------------------------------
bg_finish:					;@ end of frame...
;@----------------------------------------------------------------------------
	stmfd sp!,{r3-r9,r11,lr}

	mov r4,#16
	add r11,r4,#224				;@ r11 = sheight
	mov r8,#0x1000				;@ tilemap offset

;@----------------------------------------------------------------------------
;@bgchrfinish2		;@ end of frame...
;@----------------------------------------------------------------------------
bglo2:
	bic r0,r4,#7

	ldr r3,=EMU_RAM
	add r6,r3,r0,lsl#3			;@ 64x32 tiles
	add r5,r6,#0x800			;@ r5 tile, r6 attrib

	sub r3,r3,#0x40				;@ Scroll RAM
	ldrb r2,[r3,r0,lsr#3]!
	ldrb r3,[r3,#0x20]
	orr r2,r2,r3,lsl#8
	mov r2,r2,lsr#2

	mov r9,#BG_GFX
	add r9,r9,r0,lsl#3
	mov r3,#33					;@ width
bgtrloop:
	and r2,r2,#0x7e
	ldrb r0,[r5,r2,lsr#1]		;@ Read from GreenBeret Tilemap RAM
	ldrb r1,[r6,r2,lsr#1]		;@ Read from GreenBeret Colormap RAM
	tst r1,#0x40
	orrne r0,r0,#0x100			;@ bit 8 of tilenum
	orr r0,r0,r1,lsl#12			;@ color bits
	and r1,r0,#0x30000
	orr r0,r0,r1,lsr#6			;@ XY flip

	orr r7,r2,r2,lsl#5			;@ move 0x40 to 0x800
	bic r7,r7,#0x17C0
	strh r0,[r7,r9]!			;@ Write to GBA Tilemap RAM
	mov r0,r0,lsr#12
	and r0,r0,#0xF
	strh r0,[r7,r8]				;@ Write to GBA Tilemap RAM 2 (extra bgr color)
	add r2,r2,#2
	subs r3,r3,#1
	bne bgtrloop

	add r4,r4,#8
	cmp r4,r11					;@ 192*4/3=256
	ble bglo2

	ldmfd sp!,{r3-r9,r11,pc}

;@----------------------------------------------------------------------------
sprDMA_do:					;@ Called from endframe. YATX
;@----------------------------------------------------------------------------
	.equ PRIORITY,	0x800		;@ 0x800=AGB OBJ priority 2
	str lr,[sp,#-4]!


	ldr r9,=SPRTILELUT
	ldrb r0,sprmemreload
	cmp r0,#0
	beq noreload
	mov r0,r9					;@ r0=destination
	mov r1,#0					;@ r1=value
	mov r2,#512					;@ 512 tile entries
	bl memset_					;@ clear lut
	strb r1,sprmemreload		;@ clear spr mem reload.
	strb r1,sprmemalloc			;@ clear spr mem alloc.
noreload:

	ldr addy,=EMU_RAM
	add addy,addy,#0x1100		;@ Source
	ldrb r0,sprBank
	and r0,r0,#0x08
	sub addy,addy,r0,lsl#5
	ldr r2,tmpoambuffer			;@ Destination

	ldrb r5,g_scaling
	cmp r5,#SCALED
	mov r6,#0x40000000			;@ 16x16 size
//	orreq r6,r6,#0x100			;@ r6= scale obj

	mov r0,#0
	ldrb r4,ystart				;@ first scanline?
	cmp r5,#UNSCALED			;@ do autoscroll
	ldreq r5,=0x00010000		;@ no scaling
	ldrne r5,=0x0000DB6C		;@ 192/224, scaling
	beq dm2
;@	bhi dm2
	mov r4,#2
;@	bne dm0
	b dm0
;@	ldr r3,AGBjoypad
;@	ands r3,r3,#0x300
;@	eornes r3,r3,#0x300
;@	bne dm0						;@ stop if L or R pressed (manual scroll)
	ldr r1,=EMU_RAM+0x1200
	ldrb r0,[r1,#5]				;@ follow sprite
	rsb r0,r0,#0x80
	adds r0,r0,r0,asr#3
	cmp r0,#16
	movmi r0,#16
	cmp r0,#48
	movpl r0,#48
	str r0,windowtop
dm0:
	ldr r0,windowtop+8
dm2:
	add r4,r4,r0
	add r4,r4,#0x10
	mov r8,#48					;@ number of sprites
	add addy,addy,r8,lsl#2		;@ Begin with the last sprite
dm11:
	ldr r7,=0x1ff
	ldr r3,[addy,#-4]!			;@ GreenBeret OBJ, r4=Tile,Attrib,Xpos,Ypos.
	rsb r0,r4,r3,lsr#24			;@ mask Y
	movs r1,r3,lsr#16
	and r1,r1,#0xFF
	subcs r1,r1,#0x100			;@ Attrib bit7
	and r1,r1,r7				;@ mask X

	mul r0,r5,r0				;@ Y scaling
	and r0,r0,#0xFF0000
	add r0,r6,r0,lsr#16			;@ size + scaling + yoffset
	orr r0,r0,r1,lsl#16

	and r1,r3,#0x3000			;@ X/Yflip
	orr r0,r0,r1,lsl#16
	str r0,[r2],#4				;@ store OBJ Atr 0,1. Xpos, ypos, flip, scale/rot, size, shape.

	and r1,r3,#0xFF
	and r0,r3,#0x4000
	orr r1,r1,r0,lsr#6
	bl VRAM_spr_16				;@ jump to spr copy, takes tile# in r1, gives new tile# in r0
ret01:
	and r0,r0,#0xFF				;@ tile mask
	mov r0,r0,lsl#2
	and r1,r3,#0x0F00			;@ color
	orr r0,r0,r1,lsl#4
	orr r0,r0,#PRIORITY			;@ priority
	strh r0,[r2],#4				;@ store OBJ Atr 2. Pattern, prio & palette.
dm9:
	subs r8,r8,#1
	bne dm11
	ldr pc,[sp],#4
dm10:
	mov r0,#0x2a0				;@ double, y=160
	str r0,[r2],#8
	b dm9

;@----------------------------------------------------------------------------
VRAM_spr_16:				;@ takes tilenumber in r1, returns new tilenumber in r0
;@----------------------------------------------------------------------------
	mov r7,#1
	ldr r0,[r9,r1,lsl#2]
	cmp r7,r0,lsr#16
	movle pc,lr					;@ allready cached
noluthit16:
	bic r7,r1,#1
	ldrb r0,sprmemalloc
	orr r0,r0,#0x20000
	str r0,[r9,r7,lsl#2]
	add r7,r7,#1
	add r0,r0,#1
	str r0,[r9,r7,lsl#2]
	bic r0,r0,#1
	and r7,r1,#1
	add r0,r0,r7
	stmfd sp!,{r0-r6,lr}

	bic r1,r1,#1

	add r3,r0,#2				;@ allways alloc 2 tiles
	bic r3,r3,#1
	strb r3,sprmemalloc
	tst r3,#0x100
	movne r3,#0xab
	strneb r3,sprmemreload

;@----------------------------------------------------------------------------
do16:
	and r0,r0,#0xfe
	bic r1,r1,#0xfe00

	ldr r3,=vrombase1			;@ r3 = bitplane 0
	ldr r3,[r3]					;@ r3 = bitplane 0
	add r3,r3,r1,lsl#7
	mov r6,#SPRITE_GFX			;@ r6=NDS SPR tileset
	add r6,r6,r0,lsl#7			;@
	ldr r1,=0x0F0F0F0F

spr1:
	ldr r0,[r3],#4				;@ read 3rd plane
	and r2,r1,r0,lsr#4
	and r0,r1,r0
	orr r0,r2,r0,lsl#4
	str r0,[r6],#4

	tst r6,#0xFC				;@ allways 2 16x16 tiles
	bne spr1

cachehit16:
	ldmfd sp!,{r0-r6,pc}

;@----------------------------------------------------------------------------
DMA0BUFPTR:			.long 0

tmpoambuffer:		.long OAM_BUFFER1
dmaoambuffer:		.long OAM_BUFFER2

oambufferready:		.long 0
pcepaletteready:	.long 0
;@----------------------------------------------------------------------------

FPSValue:
	.long 0
AGBinput:			;@ this label here for main.c to use
	.long 0			;@ AGBjoypad (why is this in gfx.s again?  um.. i forget)
EMUinput:
	.long 0			;@ EMUjoypad (this is what Emu sees)
adjustblend:
	.long 1
windowtop:
	.long 0
wtop:
	.long 0,0,0		;@ windowtop  (this label too)   L/R scrolling in unscaled mode
scanline:
	.long 0
frametotal:			;@ let ui.c see frame count for savestates
	.long 0

gfxstate:
irqControl:
	.byte 0
irqPending:
	.byte 0
sprBank:
	.byte 0
sprmemalloc:
	.byte 0
sprmemreload:
	.byte 0
	.byte 0,0,0

	.pool
	.section .bss
SCRL_TMP:
	.space 0x40
SPRTILELUT:
	.space 0x800
OAM_BUFFER1:
	.space 0x400
OAM_BUFFER2:
	.space 0x400
DMA0BUFF:
	.space 0x200
SCROLLBUFF:
	.space 0x800				;@ Scrollbuffer.
MAPPED_RGB:
	.space 0x1000
EMUPALBUFF:
	.space 0x400
;@...update load/savestate if you move things around in here
;@----------------------------------------------------------------------------

